{ ModR/M Tables
 (C) 2005-2012, Andreas Berg. All Rights Reserved. }
unit _modrm;

interface

uses
 _common;

type
 TModRMTable = record
    Name           : string;
    ColCount       : TAsmValue;
    Size           : TAsmValue;
 end;
 TModRMData = record
    ModTable       : TDefValue;
    CodeOffset     : TDefValue;
    MemBits        : TAsmValue;
    BracketCode    : byte;
    Reg1           : string;
    Reg2           : string;
    Reg3           : string;
 end;

 TModRMInput = record
    ModTable       : string;
    MemBits        : TAsmValue;
    BracketCode    : byte;
    Reg            : string;
    Reg1           : string;
    Reg2           : string;
    Reg3           : string;
    Digit          : TAsmValue;
    Value          : TAsmValue;
    ReturnValue    : TAsmValue;
 end;

var
 ModRMList         : array of TModRMTable;
 ModRMCount        : TDefValue;

 ModRMDataList     : array of TModRMData;
 ModRMDataCount    : TDefValue;

 ModRMInput        : TModRMInput;

procedure InitModRM;
procedure ResetModRMInput;
function GetModTable(n: string): TDefValue;

procedure AddModRMTable(n: string; cnt: TAsmValue);
procedure AddModRMData(modTable: string);

function GetModRM(modTable: string; mbits: TAsmValue;
 brCode: byte; r1, r2, r3, reg: string): TAsmValue;
function GetModRMDigit(modTable: string; mbits: TAsmValue;
 brCode: byte; r1, r2, r3: string; digit: TAsmValue): TAsmValue;
function GetModRMSpecial(modTable: string; val: TAsmValue; reg: string): TAsmValue;

implementation

uses
 SysUtils, _error, _mem, _scanner, _parser, _reg;

procedure InitModRM;
begin
   SetLength(ModRMList, 1);
   ModRMCount:= 0;
   SetLength(ModRMDataList, 1);
   ModRMDataCount:= 0;
   ResetModRMInput;
end;

procedure ResetModRMInput;
begin
   ModRMInput.ModTable:= '';
   ModRMInput.MemBits:= 0;
   ModRMInput.BracketCode:= 0;
   ModRMInput.Reg:= '';
   ModRMInput.Reg1:= '';
   ModRMInput.Reg2:= '';
   ModRMInput.Reg3:= '';
   ModRMInput.Digit:= 0;
   ModRMInput.Value:= 0;
   ModRMInput.ReturnValue:= 0;
end;

function GetModTable(n: string): TDefValue;
var
 mTbl, i           : TDefValue;
begin
   mTbl:= Neg;
   i:= 0;
   while i < ModRMCount do begin
      if n = ModRMList[i].Name then begin
         mTbl:= i;
         break;
      end;
      Inc(i);
   end;
   GetModTable:= mTbl;
end;

procedure AddModRMTable(n: string; cnt: TAsmValue);
begin
   if GetModTable(n) = Neg then begin
      SetLength(ModRMList, ModRMCount + 1);
      ModRMList[ModRMCount].Name:= n;
      ModRMList[ModRMCount].ColCount:= cnt;
      ModRMList[ModRMCount].Size:= 0;
      Inc(ModRMCount);
   end else begin
      ErrorMsg('ModR/M Table already exists [' + n + ']!');
   end;
end;

procedure AddModRMData(modTable: string);
var
 tblIndex, i       : TDefValue;
 numCols, v        : TAsmValue;
 memOfs            : TDefValue;
 mbits             : TAsmValue;
 rg1, rg2, rg3     : string;
 brCode            : byte;
begin
   tblIndex:= GetModTable(modTable);
   if tblIndex <> Neg then begin
      numCols:= ModRMList[tblIndex].ColCount;
      memOfs:= MemLength;
      rg1:= GetName;
      rg2:= '';
      rg3:= '';
      if not IsErr then begin
         if rg1 <> '' then begin
            SkipBlank;
            if GetSrcChar = '+' then begin
               NextChar;
               rg2:= GetName;
               if not IsErr then begin
                  if rg2 <> '' then begin
                     SkipBlank;
                     if GetSrcChar = '+' then begin
                        NextChar;
                        rg3:= GetName;
                        if not IsErr then begin
                           if rg3 = '' then
                            ErrorMsg('Register name or [MEMORY] expected!');
                        end;
                     end;
                  end else begin
                     ErrorMsg('Register name or [MEMORY] expected!');
                  end;
               end;
            end;
         end else begin
            ErrorMsg('Register name or [MEMORY] expected!');
         end;
      end;

      if not IsErr then begin
         if rg1 <> '' then begin
            if rg1 <> 'MEMORY' then begin
               if GetRegIndex(rg1) = NotFound then begin
                  ErrorMsg('Register not found [' + rg1 + ']!');
               end;
            end;
            if not IsErr then begin
               if rg2 <> '' then begin
                  if rg2 <> 'MEMORY' then begin
                     if GetRegIndex(rg2) = NotFound then begin
                        ErrorMsg('Register not found [' + rg2 + ']!');
                     end;
                  end;
                  if not IsErr then begin
                     if rg3 <> '' then begin
                        if rg3 <> 'MEMORY' then begin
                           if GetRegIndex(rg3) = NotFound then begin
                              ErrorMsg('Register not found [' + rg3 + ']!');
                           end;
                        end;
                     end;
                  end;
               end;
            end;
         end;
      end;

      mbits:= GetNumber(tAUTO);

      brCode:= 0;
      if not IsErr then brCode:= GetNumber(tBYTE);
      if brCode > 6 then
       ErrorMsg('Invalid bracket code [0 to 6]!');

      if not IsErr then begin
         i:= 0;
         while i < numCols do begin
            v:= GetNumber(tAUTO);
            if IsErr then break;
            MemReplacePos:= -1;
            AddMemValue(v);
            Inc(i);
         end;
      end;

      if not IsErr then begin
         SetLength(ModRMDataList, ModRMDataCount + 1);
         ModRMDataList[ModRMDataCount].ModTable:= tblIndex;
         ModRMDataList[ModRMDataCount].CodeOffset:= memOfs;
         ModRMDataList[ModRMDataCount].MemBits:= mbits;
         ModRMDataList[ModRMDataCount].BracketCode:= brCode;
         ModRMDataList[ModRMDataCount].Reg1:= rg1;
         ModRMDataList[ModRMDataCount].Reg2:= rg2;
         ModRMDataList[ModRMDataCount].Reg3:= rg3;
         Inc(ModRMDataCount);
         ModRMList[tblIndex].Size:=
          ModRMList[tblIndex].Size + numCols;
      end;
   end else begin
      ErrorMsg('ModR/M table not found [' + modTable + ']!');
   end;
end;



function GetModRM(modTable: string; mbits: TAsmValue;
 brCode: byte; r1, r2, r3, reg: string): TAsmValue;
var
 mVal, regDigit    : TAsmValue;
 mTbl, rIndex      : TDefValue;
begin
   mVal:= 0;
   mTbl:= GetModTable(modTable);
   if mTbl <> Neg then begin
      {regDigit:= Neg;}
      rIndex:= GetRegIndex(reg);
      if rIndex <> Neg then begin
         regDigit:= Regs[rIndex].Code;
         mVal:= GetModRMDigit(modTable, mbits, brCode, r1, r2, r3, regDigit);
      end else begin
         ErrorMsg('Register not found [' + reg + ']!');
      end;
   end else begin
      ErrorMsg('ModR/M table not found [' + modTable + ']!');
   end;
   GetModRM:= mVal;
end;

function GetModRMDigit(modTable: string; mbits: TAsmValue;
 brCode: byte; r1, r2, r3: string; digit: TAsmValue): TAsmValue;
var
 mVal, cols        : TAsmValue;
 mTbl, i           : TDefValue;
 rIndex1           : TDefValue;
 rIndex2           : TDefValue;
 rIndex3           : TDefValue;
 dtModTbl          : TDefValue;
 dtCodeOfs         : TDefValue;
 dtBracket         : byte;
 dtReg1            : string;
 dtReg2            : string;
 dtReg3            : string;
begin
   mVal:= 0;
   mTbl:= GetModTable(modTable);
   if mTbl <> Neg then begin
      cols:= ModRMList[mTbl].ColCount;
      if digit >= cols then begin
         ErrorMsg('Invalid ModR/M table column (digit too high) [' +
          IntToStr(digit) + ']!');
      end else begin
         rIndex1:= Neg;
         rIndex2:= Neg;
         rIndex3:= Neg;
         if r1 <> '' then rIndex1:= GetRegIndex(r1);
         if r2 <> '' then rIndex2:= GetRegIndex(r2);
         if r3 <> '' then rIndex3:= GetRegIndex(r3);
         i:= 0;
         while i < ModRMDataCount do begin
            dtModTbl:= ModRMDataList[i].ModTable;
            dtCodeOfs:= ModRMDataList[i].CodeOffset;
            dtBracket:= ModRMDataList[i].BracketCode;
            dtReg1:= ModRMDataList[i].Reg1;
            dtReg2:= ModRMDataList[i].Reg2;
            dtReg3:= ModRMDataList[i].Reg3;
            if dtModTbl = mTbl then begin
               if (brCode = dtBracket) and (mbits = ModRMDataList[i].MemBits)
               and (rIndex1 = GetRegIndex(dtReg1)) and
                (rIndex2 = GetRegIndex(dtReg2)) and
                (rIndex3 = GetRegIndex(dtReg3)) then begin
                  mVal:= MemData[dtCodeOfs + digit];
                  break;
               end;
            end;
            Inc(i);
         end;
      end;
   end else begin
      ErrorMsg('ModR/M table not found [' + modTable + ']!');
   end;
   GetModRMDigit:= mVal;
end;

function GetModRMSpecial(modTable: string; val: TAsmValue; reg: string): TAsmValue;
var
 mVal, rCode       : TAsmValue;
 mTbl, rIndx       : TDefValue;
 cols              : TAsmValue;
begin
   mVal:= 0;
   mTbl:= GetModTable(modTable);
   if mTbl <> Neg then begin
      rIndx:= GetRegIndex(reg);
      if rIndx <> Neg then begin
         rCode:= Regs[rIndx].Code;
         cols:= ModRMList[mTbl].ColCount;
         if rCode < cols then begin
            mVal:= val + (rCode * cols);
	       end else begin
            ErrorMsg('Invalid ModR/M table column (register code too high) [' +
             reg + ']!');
         end;
      end else begin
         ErrorMsg('Register not found [' + reg + ']');
      end;
   end else begin
      ErrorMsg('ModR/M table not found [' + modTable + ']!');
   end;
   GetModRMSpecial:= mVal;
end;

end.
